@@ -1,6 +1,8 @@
+# @(#)TSM.pm	1.11
+
 package DBD::TSM;
 
-use 5.008003;
+use 5.008;
 use strict;
 #use warnings;
 
@@ -12,16 +14,16 @@ our @ISA = qw(Exporter);
 # names by default without a very good reason. Use EXPORT_OK instead.
 # Do not simply export all your public functions/methods/constants.
 
-# This allows declaration	use DBD::TSM ':all';
+# This allows declaration    use DBD::TSM ':all';
 # If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
 # will save memory.
 our %EXPORT_TAGS = ( 'all' => [ qw(
-	
+
 ) ] );
 
 our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
 
-our @EXPORT = qw(	
+our @EXPORT = qw(
 );
 
 # Preloaded methods go here.
@@ -33,7 +35,7 @@ use Carp;
 
 our ($VERSION,$err,$errstr,$sqlstate,$drh);
 
-$VERSION = '0.04';
+$VERSION = '0.11';
 ## Error in Makefile.PL, see change file
 
 # Gestion des erreurs DBI
@@ -60,12 +62,11 @@ sub driver {
                                }
                          );
 
-    # Gestion de l'erreur  la cration                    
+    # Gestion de l'erreur  la cration
     croak 'DBD::TSM: Error - Could not load driver: ',$DBI::errstr,"\n" unless $drh;
 
     # Gestion des variables d'environnement et autres prsence de l'environnement minimum
     # Fin
-
     return $drh;
 }
 
@@ -75,7 +76,6 @@ sub driver {
 #--------------------------------------------------------------
 package DBD::TSM::dr;
 
-
 use constant DEBUG => 0;
 BEGIN {
     DEBUG && require Data::Dumper;
@@ -97,39 +97,36 @@ sub disconnect_all {
 
 sub data_sources {
     my ($drh)=@_;
-    
+
     # Recuperer les infos d'un fichier
-    my @data_sources=tsm_data_sources($drh);
+    return tsm_data_sources($drh);
     # Fin
-    
-    return @data_sources;
 }
 
 sub connect {
     my ($drh, $dbname, $user, $auth, $attr) = @_;
-    
-    DEBUG && print "DEBUG - ",__PACKAGE__,"->connect: @_\n";
+
+    DEBUG && warn "DEBUG - ",__PACKAGE__,"->connect: @_\n";
 
     my $dbh = DBI::_new_dbh($drh, {
-                   'Name'         => $dbname,
-                   'USER'         => $user,
-                   'CURRENT_USER' => $user,
+                   Name         => $dbname,
+                   USER         => $user,
+                   CURRENT_USER => $user,
                });
 
     foreach my $attr_name (qw(PrintError AutoCommit RaiseError)) {
 #    foreach my $attr_name (qw(Active PrintError AutoCommit RaiseError)) {
-        my $attr_value=(exists $attr->{$attr_name})?($attr->{$attr_name}):(1);
+        my $attr_value = (exists $attr->{$attr_name}) ? $attr->{$attr_name} : 1;
         $dbh->STORE($attr_name => $attr_value);
     }
 
-    # Gestion de la connexion, c'est a dire verification que les user/password permet de se connecter
-    if (tsm_connect($dbh,$dbname,$user,$auth)) {
-        return $dbh;
-    } else {
-        return;
-    }        
-    # Fin
+    $dbh->STORE(tsm_pipe => ($^O =~ m/win/i)?(0):(1));
 
+    # Gestion de la connexion, c'est a dire verification que les user/password permet de se connecter
+    tsm_connect($dbh,$dbname,$user,$auth) and return $dbh;
+
+    return;
+    # Fin
 }
 
 #--------------------------------------------------------------
@@ -162,7 +159,8 @@ sub prepare {
              'Statement' => $statement,
              });
     $sth->STORE('NUM_OF_PARAMS' => ($statement =~ tr/\?//));
-    $sth->STORE('tsm_params'   => []);
+    $sth->STORE('tsm_params'    => []);
+    $sth->STORE('tsm_pipe'      => $dbh->{tsm_pipe});
 
     # Compilation de requete
     # Fin
@@ -225,8 +223,8 @@ sub FETCH {
 
 sub DESTROY {
     my ($dbh) = @_;
-    
-    DEBUG && print "DEBUG - ",__PACKAGE__,"->DESTROY: call @_\n";
+
+    DEBUG && warn "DEBUG - ",__PACKAGE__,"->DESTROY: call @_\n";
 }
 
 #---------------------------------------------------------------------
@@ -238,6 +236,7 @@ package DBD::TSM::st;
 use strict;
 
 use DBD::TSM::Functions;
+use Data::Dumper;
 
 use constant DEBUG => 0;
 
@@ -245,88 +244,108 @@ our $imp_data_size = 0;
 
 sub bind_param {
     my ($sth, $pNum, $val, $attr) = @_;
+    
     my $type = (ref $attr) ? $attr->{TYPE} : $attr;
+    
     if ($type) {
         my $dbh = $sth->{Database};
         $val = $dbh->quote($sth, $type);
     }
     my $params = $sth->FETCH('tsm_params');
     $params->[$pNum-1] = $val;
-    1;
+    return 1;
 }
 
 sub execute {
     my ($sth, @bind_values) = @_;
-    
+
     #Rfrence sur les paramtres d'excute
-    my $params = (@bind_values) ? \@bind_values : $sth->FETCH('tsm_params');
-    my $numOfParam=$sth->FETCH('NUM_OF_PARAMS');
-    my $num_param =scalar(@{$params});
+    $sth->finish() if ($sth->{Active});
+    
+    my $params_ref    = (@bind_values) ? \@bind_values : $sth->FETCH('tsm_params');
+    my $num_of_param  = $sth->FETCH('NUM_OF_PARAMS');
+    my $num_param     = scalar @{$params_ref};
 
     # Nombre de paramtre au moment du prepare
-    if ($numOfParam > $num_param) {
-            $sth->DBI::set_err(1,"Bad numbers of param: $num_param/$numOfParam.");
-            return;
+    if ($num_of_param > $num_param) {
+        $sth->set_err(1,"Wrong number of parameters: $num_param <> expected: $num_of_param.");
+        return;
     }
 
+    # Substitute character ? with parameters
     my $statement = $sth->{Statement};
-    my $eval_statement="";
-
-    foreach my $param (@{$params}) {
-        if (my ($sql,$next) = ($statement =~ m/([^\?]+)\?(.*)/)) {
-            $eval_statement = $eval_statement.$sql.$param;
-            $statement      = $next;
-        }
+    foreach my $param_value (@{$params_ref}) {
+        $statement =~ s/ [?] /$param_value/xms; # Substitute ? from beginning
+                                                # Check is realized by dsmadmc
     }
-    $statement = $eval_statement.$statement;
 
-    DEBUG && print "DEBUG - ",__PACKAGE__,"->execute: AutoCommit=",$sth->FETCH('AutoCommit'),"\n";
-    my ($data,$fields)=tsm_execute($sth,$statement) or return;
+    DEBUG && warn "DEBUG - ",__PACKAGE__,"->execute: AutoCommit=",$sth->FETCH('AutoCommit'),"\n";
+    my ($data_ref, $fields_ref, $rawdata_ref) = tsm_execute($sth, $statement)
+                                                or return;
 
-    my ($fields_lc,$fields_uc);
+    my ($fields_lc_ref, $fields_uc_ref);
+    @{$fields_uc_ref} = map { uc($_) } @{$fields_ref};
+    @{$fields_lc_ref} = map { lc($_) } @{$fields_ref};
 
-    @{$fields_uc}=map {uc($_)} @{$fields};
-    @{$fields_lc}=map {lc($_)} @{$fields};
+    # Store parameters
+    $sth->STORE(tsm_data        => $data_ref);
+    $sth->STORE(tsm_raw         => $rawdata_ref);
+    $sth->STORE(tsm_rows        => scalar @{$data_ref}); # number of rows
 
-    # Stock les parametres
-    $sth->STORE('tsm_data',$data);
-    $sth->STORE('tsm_rows',scalar(@{$data})); # number of rows
-    my $nb_fields=@{$fields};
-    DEBUG && print "DEBUG - ",__PACKAGE__,"->execute: nb fields = $nb_fields, ",$sth->FETCH('NUM_OF_FIELDS'),"\n";
-    $sth->STORE('NUM_OF_FIELDS',$nb_fields); # unless ($nb_fields == $sth->FETCH('NUM_OF_FIELDS')); pourquoi je faisais ce test?
-    $sth->STORE('NAME',$fields);
-    $sth->STORE('NAME_lc',$fields_lc);
-    $sth->STORE('NAME_uc',$fields_uc);
-    $sth->STORE('NULLABLE'      => [ (0) x @{$fields} ]);
-    $sth->STORE('TYPE'          => [ (DBI::SQL_VARCHAR()) x @{$fields} ]);
-    $sth->STORE('SCALE'         => undef);
-    $sth->STORE('PRECISION'     => undef);
+    #Number of fields, already set by other routine ?
+    my $nb_fields = @{$fields_ref};
+    DEBUG && warn   "DEBUG - ", __PACKAGE__
+                  , "->execute: nb fields = $nb_fields, "
+                  , $sth->FETCH('NUM_OF_FIELDS')
+                  , "\n";
+    $sth->STORE(NUM_OF_FIELDS => $nb_fields) unless (
+                                                 $sth->FETCH('NUM_OF_FIELDS')
+                                             and $nb_fields == $sth->FETCH('NUM_OF_FIELDS')
+                                             ); #pourquoi faut il faire ce test?
+    $sth->STORE(NAME          => $fields_ref);
+    $sth->STORE(NAME_lc       => $fields_lc_ref);
+    $sth->STORE(NAME_uc       => $fields_uc_ref);
+    $sth->STORE(NULLABLE      => [ (0) x @{$fields_ref} ]);
+    $sth->STORE(TYPE          => [ (DBI::SQL_VARCHAR()) x @{$fields_ref} ]);
+    $sth->STORE(SCALE         => undef);
+    $sth->STORE(PRECISION     => undef);
+    
+    DEBUG && warn "DEBUG:Execute: ", Dumper($data_ref);
 
-    return (@{$data} || '0E0');
+    return (@{$data_ref} || '0E0');
 }
 
 sub fetchrow_arrayref {
      my ($sth) = @_;
 
-     my $data = $sth->FETCH('tsm_data');
-     my $row  = shift @{$data};
-
+     my $data_ref = $sth->FETCH('tsm_data');
+     my $row_ref  = shift @{$data_ref};
+     
+     DEBUG && warn "DEBUG:Line: ", Dumper($row_ref);
+     
      # Fin du tableau
-     return undef unless ($row);
+     unless ($row_ref) {
+        DEBUG && "DEBUG:Line: Fini.\n";
+        $sth->{Active} = 0;
+        return undef;
+     }
 
      if ($sth->FETCH('ChopBlanks')) {
-        map { $_ =~ s/\s+$//;
-              $_ =~ s/^\s+//;
-             } @$row;
+        foreach (@{$row_ref}) {
+            s/\s+$//;
+            s/^\s+//;
+        }
      }
-     return $sth->_set_fbav($row);
+
+     return $sth->_set_fbav($row_ref);
 }
 
 *fetch = \&fetchrow_arrayref;
 
 sub rows {
     my ($sth) = @_;
-    return ($sth->FETCH('tsm_rows'));
+
+    return $sth->FETCH('tsm_rows');
 }
 
 sub STORE {
@@ -334,7 +353,7 @@ sub STORE {
 
     return 1 if ($attr eq 'AutoCommit');
 
-    if ($attr =~ m/^tsm_/  or
+    if ($attr =~ m/^tsm_/   or
         $attr =~ m/^NAME/   or
         $attr eq 'NULLABLE' or
         $attr eq 'SCALE'    or
@@ -369,7 +388,7 @@ DBD::TSM - Perl DBD driver for TSM admin client
 =head1 SYNOPSIS
 
     use DBI;
-    
+
     my ($server_name,$user,$password)=(...); #or set DBI_DSN,DBI_USER,DBI_PASS
 
     my $dbh=DBI->connect("DBB:TSM:$server_name",$user,$password,
@@ -377,7 +396,7 @@ DBD::TSM - Perl DBD driver for TSM admin client
                          PrintError => 0}) or die $DBI::errstr;
     #If you use environment variable $dbh=DBI->connect();
 
-    my $sth=$dbh->prepare("select node_name from nodes") or 
+    my $sth=$dbh->prepare("select node_name from nodes") or
             die $dbh->errstr;
     $sth->execute() or die $sth->errstr();
 
@@ -436,13 +455,24 @@ Check it before with a manual test with:
 
 When you set AutoCommit (the default), I add -itemcommit in command line.
 
-=head2 Advice
+=head2 Exception
 
-I always set PrintError => 0 and RaiseError => 0. Because, TSM send a '11' return code
-for empty statement and DBI exit automatically with rc > 0.
+You have to ways to track execute error in script. It's mandatory with TSM because, it
+send a 11 return code for empty statement. I propagate this return code.
 
-I prefer to check return code with function err and messages with errmsg. I do a filter 
-on AN.....[^I] to print error messages.
+So you have two methods to not exit from script:
+
+1. First, set RaiseError => 0 in connect method
+
+    my $dbh = DBI->connect($dbi_dsn, $dbi_user, $dbi_pass, {
+            RaiseError => 0,
+        });
+
+2. Use eval {}; block for execute fonction
+
+    eval {
+        $sth->execute($select);
+    };
 
 =head2 EXPORT
 
@@ -456,12 +486,19 @@ TSM Client Reference Manual.
 
 =head1 BUGS
 
-I'm not using TSM API. So, I do one session for each statement. I'm not a C developper, 
-sor it's a Pure Perl Module.
+I'm not using TSM API. So, I do one session for each statement. It's a Pure Perl Module.
+
+Be carefull with join statement because we could have duplicate field name. I detect this duplicate
+field and send a warning message.
+
+=head1 NEXT
+
+Rewrite fetch to use a filehandle to read dsmadmc output line by line to redure memory 
+requirement. I have some idea, need just time to do it. Not sure it works on Windows.
 
 =head1 AUTHOR
 
-Laurent Bendavid, E<lt>bendavid.laurent@free.frE<gt>
+Laurent Bendavid, E<lt>lbendavid@cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
